class BadRequest(Exception):
    status_code = 400

class Forbidden(BadRequest):
    status_code = 403

class NotFound(BadRequest):
    status_code = 404

class NotAllowed(BadRequest):
    status_code = 405

class ImATeapot(BadRequest):
    status_code = 418



from django.shortcuts import render, reverse
from django.http import JsonResponse

def describe_route(urlpattern):
    """Introspects data from a url pattern"""
    p = urlpattern
    f = p.callback
    return {
        "name": p.name,
        "human_name": p.name.replace("_", " ").capitalize(),
        "description": f.__doc__ or "",
        "arguments": {param: type_.__name__ for param, type_ in f.args.items()},
        "url": reverse(p.name),
        "api_url": reverse("api_" + p.name),
        "method": f.method,
    }

def get_routes():
    """Returns a list of dicts, each describing a route in the app.
    """
    from banjo.urls import user_defined_routes
    return [describe_route(r) for r in user_defined_routes]

def api(request):
    """A view which renders the API index page.
    """
    return render(request, "banjo/api.html", {"routes": get_routes()})
    
def api_json(request):
    "Returns a JSON representation of available API routes"
    routes = {'routes': get_routes()}
    return JsonResponse(routes)



import os
from os.path import abspath, dirname, join
import sys
import django
import shutil
from pathlib import Path
from subprocess import run
from django.conf import settings
from django.core import management

sys.path.insert(0, '.')

def setup_django(debug=False, settings=None):
    """Sets up Django and runs prerequisite management commands.

    Django needs to be configured before any app-specific code can be loaded.
    Sets the environment variable ``DJANGO_SETTINGS_MODULE`` so Django uses one of 
    Banjo's settings modules (and so the user's app doesn't need a settings file). 
    Then runs ``django.setup()``. 

    Imports the app's views module, which 
    causes banjo's route_get and route_post decorators to be invoked, 
    which populates ``banjo.urls:urlpatterns``.

    Finally, executes Django management commands ``makemigrations`` and ``migrate``, 
    ensuring that the database is in sync with the app's models.
    """
    if settings:
        os.environ.setdefault('DJANGO_SETTINGS_MODULE', settings)
    elif debug:
        os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'banjo.settings_debug')
    else:
        os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'banjo.settings')
    django.setup()

    from app import views

    management.execute_from_command_line(['', 'makemigrations', 'app', 'banjo'])
    management.execute_from_command_line(['', 'migrate'])

def main():
    """The entry point for the Banjo CLI. 
    Declares arguments and then invokes Django management commands. If ``--shell``, 
    invokes ``shell_plus``, otherwise invokes ``runserver``, passing along 
    ``--port`` and ``--debug`` options.
    """
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("-s", "--shell", action="store_true")
    parser.add_argument("-p", "--port", type=int, default=8000)
    parser.add_argument("-d", "--debug", action="store_true")
    parser.add_argument("-m", "--settings")
    args = parser.parse_args()

    if args.shell:
        setup_django(args.debug, settings=args.settings)
        management.execute_from_command_line(['', 'shell_plus'])
    else:
        setup_django(args.debug, settings=args.settings)
        management.execute_from_command_line(['', 'runserver', str(args.port)])



from django.urls import path
from django.http import JsonResponse, HttpResponseNotAllowed
from django.shortcuts import render
from banjo import http
from banjo import views
from banjo.forms import ApiRouteForm
import json

urlpatterns = [
    path("api", views.api, name="api"),
    path("api.json", views.api_json, name="api_json"),
]
user_defined_routes = []

def validate_args(args):
    """Checks that args (the param type signature) is valid.
    """
    if args is None: return
    allowed_types = [str, bool, int, float]
    if not isinstance(args, dict):
        raise ValueError("args must be a dict if provided")
    for param, type_ in args.items():
        if type_ not in allowed_types:
            raise ValueError("param {} in args must be one of: {}".format(
                    param, ', '.join(str(t) for t in allowed_types)))

def create_view(fn, method, args=None):
    """Creates a django view function.
    The view function checks that the HTTP method is correct, 
    extracts the request's params, passes them to ``fn``, 
    and returns the result. Args, if provided, should be 
    a dict of ``{param_name: type}``, where type is in 
    ``[str, bool, int, float]``.
    """
    method = method.upper()
    def view(request):
        if request.method != method:
            return HttpResponseNotAllowed([method])  
        if args:
            if request.headers.get('Content-Type') == 'application/json':
                data = json.loads(request.body.decode("utf-8"))
            elif method == 'GET':
                data = request.GET
            else:
                data = request.POST
            FormClass = ApiRouteForm.for_args(args)
            form = FormClass(data)
            if form.is_valid():
                try:
                    result = fn(form.cleaned_data)
                    return JsonResponse(result)
                except http.BadRequest as e:
                    return JsonResponse({'error': str(e)}, status=e.status_code)
            else:
                return JsonResponse({'errors': form.errors}, status=400)
        else:
            try:
                result = fn({})
                return JsonResponse(result)
            except http.BadRequest as e:
                return JsonResponse({'error': str(e)}, status=e.status_code)

    view.__name__ = fn.__name__
    view.__doc__ = fn.__doc__
    view.method = method
    validate_args(args)
    view.args = args or {}
    return view

def create_api_view(url, fn, method, args=None):
    """Creates an API view.
    An API view responds to GET requests with a HTML template--
    the response data for GET routes and a form for POST routes.
    For POST routes, the view processes the form.
    """
    method = method.upper()
    fn.method = method
    def api_view(request):
        fn.args = args or {}
        route = views.describe_route(path(url, fn, name=fn.__name__))

        if method == "GET" and request.method == "GET":
            if args:
                form = ApiRouteForm.for_args(args)(request.GET)
                if form.is_valid():
                    try:
                        result = fn(form.cleaned_data)
                    except http.BadRequest as e:
                        result = {'error': str(e), 'status_code': e.status_code}
                else:
                    result = None
            else:
                form = None
                try:
                    result = fn({})
                except http.BadRequest as e:
                    result = {'error': str(e), 'status_code': e.status_code}
            return render(request, "banjo/api_get.html", {
                "route": route,
                "result": json.dumps(result, indent=True) if result else None,
                "form": form,
            })
        if method == "POST":
            if request.method == "POST":
                if args:
                    form = ApiRouteForm.for_args(args)(request.POST)
                    if form.is_valid():
                        try:
                            result = fn(form.cleaned_data)
                        except http.BadRequest as e:
                            result = {'error': str(e), 'status_code': e.status_code}
                        form = ApiRouteForm.for_args(args)()
                    else:
                        result = None
                else:
                    form = None
                    try:
                        result = fn({})
                    except http.BadRequest as e:
                        result = {'error': str(e), 'status_code': e.status_code}
            else:
                if args:
                    form = ApiRouteForm.for_args(args)()
                else:
                    form = None
                result = None

            return render(request, "banjo/api_post.html", {
                "route": route,
                "result": json.dumps(result, indent=True) if result else None,
                "form": form,
            })

        return HttpResponseNotAllowed([request.method])

    api_view.__name__ = fn.__name__
    api_view.__doc__ = fn.__doc__
    validate_args(args)
    api_view.args = args or {}
    return api_view
    

def route_get(url, args=None):
    """A decorator which registers a HTTP GET route in the Banjo app.
    """
    def bind_url_to_view(fn):
        view = create_view(fn, "GET", args=args)
        api_view = create_api_view(url, fn, "GET", args=args)
        urlpatterns.append(path(url, view, name=view.__name__))
        urlpatterns.append(path("api/" + url, api_view, name="api_" + view.__name__))
        user_defined_routes.append(path(url, view, name=view.__name__))
        return view
    return bind_url_to_view

def route_post(url, args=None):
    """A decorator which registers a HTTP POST route in the Banjo app.
    """
    def bind_url_to_view(fn):
        view = create_view(fn, "POST", args=args)
        api_view = create_api_view(url, fn, "POST", args=args)
        urlpatterns.append(path(url, view, name=view.__name__))
        urlpatterns.append(path("api/" + url, api_view, name="api_" + view.__name__))
        user_defined_routes.append(path(url, view, name=view.__name__))
        return view
    return bind_url_to_view



from django import forms

type_fields = {
    str: lambda: forms.CharField(),
    bool: lambda: forms.BooleanField(required=False),
    int: lambda: forms.IntegerField(),
    float: lambda: forms.FloatField(),
}

class ApiRouteForm(forms.Form):
    """Dynamically instantiates a form matching a route's args dict.
    ApiRouteForm is used in automatically constructing the API page for
    each route. Args should be a dict of {param_name: type}, where type is in 
    [str, bool, int, float].
    """

    @classmethod
    def for_args(cls, args, class_name=None):
        "Dynamically define a form class for the given route args."
        class_name = class_name or "ApiRouteFormWithArgs"
        fields = {name: type_fields[t]() for name, t in args.items()}
        return type(class_name, (ApiRouteForm,), fields)



from os.path import abspath, dirname, join

BASE_DIR = dirname(abspath(__file__))
DEBUG=False
ROOT_URLCONF = "banjo.urls"
DATABASES = {
    "default": {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'database.sqlite',
    }
}
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]
ALLOWED_HOSTS = ["*"]
SECRET_KEY = "xxx"
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    "django_extensions",
    "banjo",
]
MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]
STATIC_URL = '/static/'
SHELL_PLUS_DONT_LOAD = [
    "admin",
    "auth",
    "contenttypes",
    "sessions",
]
SHELL_PLUS_DJANGO_IMPORTS = False



from django.db import models
from random import choice, sample

class RandomItemQuerySet(models.QuerySet):
    """Extends the base QuerySey with random() and sample() methods
    """
    def random(self):
        ids = self.values_list('id', flat=True)
        if not ids:
            raise self.model.DoesNotExist(f"{self.model.__name__} is empty")
        return self.get(id=choice(ids))

    def sample(self, n):
        ids = self.values_list('id', flat=True)
        if len(ids) < n:
            raise self.model.DoesNotExist(f"Requested {n} but {self.model.__name__} only contains {len(ids)} objects.")
        return self.filter(id__in=sample([i for i in ids], n))

class Model(models.Model):

    objects = models.Manager.from_queryset(RandomItemQuerySet)()

    @classmethod
    def from_dict(cls, props):
        """Tries to create an instance of Model from props.
        """
        if 'id' in props:
            del props['id']
        return cls(**props)

    def to_dict(self):
        """Returns a json representation of the Model.
        """
        field_names = [f.name for f in self._meta.get_fields() if f.name != "model_ptr"]
        d = {}
        for name in field_names:
            value = getattr(self, name)
            if isinstance(value, Model):
                d[name] = {'id': value.id}
            elif hasattr(value, 'all'):
                d[name] = [{'id': obj.id} for obj in value.all()]
            else:
                d[name] = value
        return d

    def __str__(self):
        return "<{} {}>".format(self.__class__.__name__, self.to_dict())

    def __repr__(self):
        return self.__str__()

    class Meta:
        abstract = True
        app_label = "app"

class BooleanField(models.BooleanField):
    """A database column which stores a boolean.
    The default value is False.
    """
    def __init__(self, *args, **kwargs):
        kwargs['default'] = False
        models.BooleanField.__init__(self, *args, **kwargs)

class IntegerField(models.IntegerField):
    """A database column which stores an integer.
    The default value is 0.
    """
    def __init__(self, *args, **kwargs):
        kwargs['default'] = 0
        models.IntegerField.__init__(self, *args, **kwargs)

class FloatField(models.FloatField):
    """A database column which stores a float.
    The default value is 0.0.
    """
    def __init__(self, *args, **kwargs):
        kwargs['default'] = 0.0
        models.FloatField.__init__(self, *args, **kwargs)

class StringField(models.TextField):
    """A database column which stores a string.
    The default value is '', the empty string.
    """
    def __init__(self, *args, **kwargs):
        kwargs['default'] = ''
        models.TextField.__init__(self, *args, **kwargs)

class ForeignKey(models.ForeignKey):
    """A database column which links a model to another model.
    """
    def __init__(self, *args, **kwargs):
        kwargs['on_delete'] = models.CASCADE
        models.ForeignKey.__init__(self, *args, **kwargs)
